# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.pydlutils.spheregroup.
"""
import pytest
import numpy as np
from astropy.utils.data import get_pkg_data_filename
from ..spheregroup import spheregroup, spherematch
from .. import PydlutilsException, PydlutilsUserWarning


def test_spheregroup():
    test_data_file = get_pkg_data_filename('t/spheregroup_data.txt')
    test_data = np.loadtxt(test_data_file, dtype='d', delimiter=',')
    # np.random.seed(137)
    # Ngroup = 3
    # N = 50
    # spread = 0.05
    linklength = 3.0  # degrees
    # x0 = np.concatenate( ( np.random.normal(loc=1,scale=spread,size=(N,)),
    #                        np.random.normal(loc=-1,scale=spread,size=(N,)),
    #                        np.random.normal(loc=1,scale=spread,size=(N,)),
    #                      )).reshape((N*Ngroup,))
    # y0 = np.concatenate( ( np.random.normal(loc=1,scale=spread,size=(N,)),
    #                        np.random.normal(loc=-1,scale=spread,size=(N,)),
    #                        np.random.normal(loc=-1,scale=spread,size=(N,)),
    #                      )).reshape((N*Ngroup,))
    # z0 = np.concatenate( ( np.random.normal(loc=1,scale=spread,size=(N,)),
    #                        np.random.normal(loc=-1,scale=spread,size=(N,)),
    #                        np.random.normal(loc=0,scale=spread,size=(N,)),
    #                      )).reshape((N*Ngroup,))
    # foo = np.arange(N*Ngroup)
    # np.random.shuffle(foo)
    # x = x0[foo]
    # y = y0[foo]
    # z = z0[foo]
    # r = np.sqrt(x**2 + y**2 + z**2)
    # theta = np.degrees(np.arccos(z/r))
    # phi = np.degrees(np.arctan2(y,x))
    # ra = np.where(phi < 0, phi + 360.0,phi)
    # dec = 90.0 - theta
    # group = spheregroup(ra,dec,linklength)
    #
    # Reproduce IDL results
    #
    ra = test_data[0, :]
    dec = test_data[1, :]
    expected_ingroup = test_data[2, :].astype(np.int64)
    expected_multgroup = test_data[3, :].astype(np.int64)
    expected_firstgroup = test_data[4, :].astype(np.int64)
    expected_nextgroup = test_data[5, :].astype(np.int64)
    group = spheregroup(ra, dec, linklength)
    assert (group[0] == expected_ingroup).all()
    assert (group[1] == expected_multgroup).all()
    assert (group[2] == expected_firstgroup).all()
    assert (group[3] == expected_nextgroup).all()
    #
    # Exceptions
    #
    with pytest.raises(PydlutilsException):
        group = spheregroup(np.array([137.0]), np.array([55.0]), linklength)
    #
    # warnings
    #
    with pytest.warns(PydlutilsUserWarning) as w:
        group = spheregroup(ra, dec, linklength, chunksize=linklength)
    # w = recwarn.pop(PydlutilsUserWarning)
    assert "chunksize changed to" in str(w[0].message)


def test_spherematch():
    i1_should_be = np.array([17, 0, 2, 16, 12, 13, 1, 5, 15, 7,
                             19, 8, 11, 10, 14, 18, 3, 9, 6, 4])
    i2_should_be = np.array([2, 0, 17, 3, 16, 15, 14, 5, 6, 10,
                             8, 19, 18, 4, 9, 7, 11, 12, 1, 13])
    np.random.seed(137)
    searchrad = 3.0/3600.0
    n = 20
    ra1 = 360.0*np.random.random((n,))
    dec1 = 90.0 - np.rad2deg(np.arccos(2.0*np.random.random((n,)) - 1.0))
    ra2 = ra1 + np.random.normal(0, 1.0/3600.0)
    dec2 = dec1 + np.random.normal(0, 1.0/3600.0)
    foo = np.arange(n)
    np.random.shuffle(foo)
    i1, i2, d12 = spherematch(ra1, dec1, ra2[foo], dec2[foo], searchrad,
                              maxmatch=0)
    assert (i1 == i1_should_be).all()
    assert (i2 == i2_should_be).all()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the window directory in photoop.
"""
import os
import time
from warnings import warn
import numpy as np
from astropy import log
from astropy.io import fits
from astropy.io.fits.fitsrec import FITS_rec
from astropy.table import Table
from . import PhotoopException
from .sdssio import sdss_name, sdss_calib
from ..pydlutils.mangle import FITS_polygon
from ..pydlutils.sdss import sdss_flagval


def sdss_score(flist, silent=True, **kwargs):
    """Score a list of imaging fields from zero to one.

    Additional keyword arguments are passed to :func:`~pydl.photoop.sdssio.sdss_calib`.

    Parameters
    ----------
    flist : :class:`~astropy.io.fits.HDUList`
        Opened FITS file, typically ``window_flist.fits``.
    silent : :class:`bool`, optional
        If ``False``, print extra information.

    Returns
    -------
    :class:`numpy.ndarray`
        A vector of scores, one for each row of the FITS file.
    """
    #
    # Not sure why these were defined.
    #
    # lat = 32.780361
    # lng = 360.0 - 105.820417
    # tzone = 7
    # scores = 1
    #
    # Read the PHOTO status bits
    #
    if not silent:
        log.info('Setting PHOTO status bits')
    t1 = time.time()
    nlist = flist[1].header.get('NAXIS2')
    fdata = flist[1].data
    for k in range(nlist):
        if not silent and ((k % 1000) == 0):
            log.info("Setting PHOTO status {0:d} {1:d}".format(k, nlist))
        thisfile = sdss_name('fpFieldStat', fdata.field('RUN')[k],
                             fdata.field('CAMCOL')[k], fdata.field('FIELD')[k],
                             fdata.field('RERUN')[k])
        try:
            fpfield = fits.open(thisfile)
        except IOError:
            warn("Retrying fpFieldStat file for RUN = {0:d} CAMCOL = {1:d} FIELD = {2:d} RERUN = {3}".format(fdata.field('RUN')[k],
                 fdata.field('CAMCOL')[k], fdata.field('FIELD')[k],
                 fdata.field('RERUN')[k]))
            try:
                fpfield = fits.open(thisfile)
            except IOError:
                warn("Bad fpFieldStat file for RUN = {0:d} CAMCOL = {1:d} FIELD = {2:d} RERUN = {3}".format(fdata.field('RUN')[k],
                     fdata.field('CAMCOL')[k], fdata.field('FIELD')[k],
                     fdata.field('RERUN')[k]))
                fdata.field('PHOTO_STATUS')[k] = -1
                if not silent:
                    log.info('Trying tsField instead.')
                thisfile = sdss_name('tsField', fdata.field('RUN')[k],
                                     fdata.field('CAMCOL')[k],
                                     fdata.field('FIELD')[k],
                                     fdata.field('RERUN')[k])
                try:
                    tsfield = fits.open(thisfile)
                except IOError:
                    warn('Bad tsField file.')
                else:
                    if not silent:
                        log.info('tsField found, using frames_status.')
                    fdata.field('PHOTO_STATUS')[k] = tsfield[1].data.field('frames_status')[0]
            else:
                fdata.field('PHOTO_STATUS')[k] = fpfield[1].data.field('status')[0]
        else:
            fdata.field('PHOTO_STATUS')[k] = fpfield[1].data.field('status')[0]
    if not silent:
        log.info("Time to set PHOTO status = {0:f} sec".format(time.time()-t1))
    #
    # Read in the PSP status
    #
    if not silent:
        log.info('Setting PSP status bits')
    t2 = time.time()
    for k in range(nlist):
        if not silent and ((k % 1000) == 0):
            log.info("Setting PSP status {0:d} {1:d}".format(k, nlist))
        thisfile = sdss_name('psField', fdata.field('RUN')[k],
                             fdata.field('CAMCOL')[k], fdata.field('FIELD')[k],
                             fdata.field('RERUN')[k])
        try:
            psfield = fits.open(thisfile)
        except IOError:
            warn("Retrying psField file for RUN = {0:d} CAMCOL = {1:d} FIELD = {2:d} RERUN = {3}".format(fdata.field('RUN')[k],
                 fdata.field('CAMCOL')[k], fdata.field('FIELD')[k],
                 fdata.field('RERUN')[k]))
            try:
                psfield = fits.open(thisfile)
            except IOError:
                warn("Bad psField file for RUN = {0:d} CAMCOL = {1:d} FIELD = {2:d} RERUN = {3}".format(fdata.field('RUN')[k],
                     fdata.field('CAMCOL')[k], fdata.field('FIELD')[k],
                     fdata.field('RERUN')[k]))
                fdata.field('PSP_STATUS')[k] = -1
                fdata.field('PSF_FWHM')[k] = -1
                fdata.field('SKYFLUX')[k] = -1
        pixscale = 0.396 * np.sqrt(fdata.field('XBIN')[k]**2 +
                                   fdata.field('YBIN')[k]**2)/np.sqrt(2.0)
        calibinfo = sdss_calib(fdata.field('RUN')[k],
                               fdata.field('CAMCOL')[k],
                               fdata.field('FIELD')[k],
                               fdata.field('RERUN')[k], **kwargs)
        fdata.field('PSP_STATUS')[k] = psfield[6].data.field('status')[0]
        fdata.field('PSF_FWHM')[k] = psfield[6].data.field('psf_width')[0]
        fdata.field('SKYFLUX')[k] = (psfield[6].data.field('sky')[0] *
                                     calibinfo['NMGYPERCOUNT']/pixscale**2)
    if not silent:
        log.info("Time to set PSP status = {0:f} sec".format(time.time()-t2))
    #
    # Decide if each field exists in all 5 bands.
    #
    bad_bits = sdss_flagval('image_status', ['bad_rotator', 'bad_astrom',
                            'bad_focus', 'shutters'])
    if 'ignoreframesstatus' in kwargs:
        ignoreframesstatus = np.zeros(fdata.field('PHOTO_STATUS').shape) == 0
    else:
        ignoreframesstatus = np.zeros(fdata.field('PHOTO_STATUS').shape) == 1
    qexist = (fdata.field('PHOTO_STATUS') == 0) | ignoreframesstatus
    for k in range(5):
        qexist &= (fdata.field('IMAGE_STATUS')[:, k] & bad_bits) == 0
    #
    # Decide if each field is phtometric in all 5 bands.
    #
    unphot_bits = sdss_flagval('image_status', ['cloudy', 'unknown',
                               'ff_petals', 'dead_ccd', 'noisy_ccd'])
    qphot = fdata.field('SUN_ANGLE') < -12
    for k in range(5):
        qphot &= (fdata.field('IMAGE_STATUS')[:, k] & unphot_bits) == 0
    for k in range(5):
        qphot &= (((fdata.field('PSP_STATUS')[:, k] & 31) <= 2) |
                  (fdata.field('XBIN') > 1) | ignoreframesstatus)
    #
    # Now set the score for each field
    #
    sensitivity = (0.7 / (fdata.field('PSF_FWHM')[:, 2] *
                   np.sqrt(fdata.field('SKYFLUX')[:, 2]))) < 0.4
    fdata.field('SCORE')[:] = qexist * (0.1 + 0.5*qphot + sensitivity)
    ibinned = np.find(fdata.field('XBIN') > 1)
    if len(ibinned) > 0:
        fdata.field('SCORE')[ibinned] *= 0.1
    #
    # Look for any NaN values, which could happen for example if there
    # is a corrupted psField file and PSF_FWHM or SKYFLUX was negative.
    #
    ibad = np.find(~np.isfinite(fdata.field('SCORE')))
    if len(ibad) > 0:
        warn("Changing NaN scores for {0:d} fields to zero.".format(len(ibad)))
        fdata.field('SCORE')[ibad] = 0
    return fdata.field('SCORE')


def window_read(flist=False, rescore=False, blist=False, bcaps=False,
                balkans=False, findx=False, bindx=False):
    """Read window files in :envvar:`PHOTO_RESOLVE`.

    Parameters
    ----------
    flist : :class:`bool`, optional
        If ``True``, read the ``window_flist.fits`` file.
    rescore : :class:`bool`, optional
        If `flist` is ``True``, look for ``window_flist_rescore.fits``, and run
        :func:`~pydl.photoop.window.window_score` if it is not found.
    blist : :class:`bool`, optional
        If ``True``, read the ``window_blist.fits`` file.
    bcaps : :class:`bool`, optional
        If ``True``, read the ``window_bcaps.fits`` file.
    balkans : :class:`bool`, optional
        If ``True``, construct the balkans from the ``window_blist.fits`` and
        ``window_bcaps.fits`` files.
    findx : :class:`bool`, optional
        If ``True``, read the ``window_findx.fits`` file.
    bindx : :class:`bool`, optional
        If ``True``, read the ``window_bindx.fits`` file.

    Returns
    -------
    :class:`dict`
        A dictionary containing the requested window data.

    Notes
    -----
    If `balkans` is ``True``, the balkans data will be in the form of a
    :class:`~pydl.pydlutils.mangle.FITS_polygon` object, to facilitate
    interoperability with :mod:`pydl.pydlutils.mangle`.  In this object,
    the keyword ``IFIELD`` is equivalent to ``IPRIMARY`` and ``PIXEL`` is
    eqivalent to ``IBINDX``.
    """
    try:
        resolve_dir = os.environ['PHOTO_RESOLVE']
    except KeyError:
        raise PhotoopException(('You have not set the environment variable ' +
                                'PHOTO_RESOLVE!'))
    r = dict()
    if flist:
        flist_file = os.path.join(resolve_dir, 'window_flist.fits')
        if rescore:
            flist_file = os.path.join(resolve_dir, 'window_flist_rescore.fits')
            if not os.path.exists(flist_file):
                window_score(rescore=rescore)
        r['flist'] = Table.read(flist_file, hdu=1)
    if blist or balkans:
        blist_file = os.path.join(resolve_dir, 'window_blist.fits')
        r['blist'] = Table.read(blist_file, hdu=1)
    if bcaps or balkans:
        bcaps_file = os.path.join(resolve_dir, 'window_bcaps.fits')
        r['bcaps'] = Table.read(bcaps_file, hdu=1)
    if findx:
        findx_file = os.path.join(resolve_dir, 'window_findx.fits')
        r['findx'] = Table.read(findx_file, hdu=1)
    if bindx:
        bindx_file = os.path.join(resolve_dir, 'window_bindx.fits')
        r['bindx'] = Table.read(bindx_file, hdu=1)
    if balkans:
        max_caps = r['blist']['NCAPS'].max()
        r['balkans'] = np.recarray((len(r['blist']),),
                                   dtype=[('IFIELD', r['blist']['IPRIMARY'].dtype),
                                          ('PIXEL', r['blist']['IBINDX'].dtype),
                                          ('NCAPS', r['blist']['NCAPS'].dtype),
                                          ('USE_CAPS', np.int32),
                                          ('WEIGHT', r['blist']['WEIGHT'].dtype),
                                          ('STR', r['blist']['STR'].dtype),
                                          ('XCAPS', r['bcaps']['X'].dtype, (max_caps, 3)),
                                          ('CMCAPS', r['bcaps']['CM'].dtype, (max_caps,)),
                                          ]).view(FITS_rec)
        r['balkans']['IFIELD'] = r['blist']['IPRIMARY']
        r['balkans']['PIXEL'] = r['blist']['IBINDX']
        r['balkans']['NCAPS'] = r['blist']['NCAPS']
        r['balkans']['USE_CAPS'] = (1 << r['blist']['NCAPS']) - 1
        r['balkans']['WEIGHT'] = r['blist']['WEIGHT']
        r['balkans']['STR'] = r['blist']['STR']
        for k in range(len(r['blist'])):
            r['balkans'][k]['XCAPS'][0:r['blist']['NCAPS'][k], :] = r['bcaps']['X'][r['blist']['ICAP'][k]:r['blist']['ICAP'][k]+r['blist']['NCAPS'][k], :]
            r['balkans'][k]['CMCAPS'][0:r['blist']['NCAPS'][k]] = r['bcaps']['CM'][r['blist']['ICAP'][k]:r['blist']['ICAP'][k]+r['blist']['NCAPS'][k]]
            #
            # Since allow_doubles is True, would set_use_caps ever return anything but the default value?
            # Comparing to the IDL code, it does appear to be the case that
            # set_use_caps is completely useless.
            #
            # r['balkans'][k]['USE_CAPS'] = set_use_caps(ManglePolygon(r['balkans'][k]),
            #                                            np.arange(r['balkans'][k]['NCAPS'], dtype=np.int32),
            #                                            allow_doubles=True)
        r['balkans'] = r['balkans'].view(FITS_polygon)
        if not blist:
            del r['blist']
        if not bcaps:
            del r['bcaps']
    return r


def window_score(rescore=False):
    """For uber-resolve, score all the fields from zero to one.

    Parameters
    ----------
    rescore : :class:`bool`, optional
        If `rescore` is ``True``, then write a new file 'window_flist_rescore.fits'
        rather than over-writing the file 'window_flist.fits'
    """
    #
    # Be certain not to use global calibrations
    #
    try:
        calib_dir_save = os.environ['PHOTO_CALIB']
    except KeyError:
        raise PhotoopException(('You have not set the environment variable ' +
                                'PHOTO_CALIB!'))
    del os.environ['PHOTO_CALIB']
    #
    # Read the file
    #
    try:
        resolve_dir = os.environ['PHOTO_RESOLVE']
    except KeyError:
        raise PhotoopException(('You have not set the environment variable ' +
                                'PHOTO_RESOLVE!'))
    filename = os.path.join(resolve_dir, 'window_flist.fits')
    if rescore:
        fitsmode = 'readonly'
    else:
        fitsmode = 'update'
    try:
        flist = fits.open(filename, mode=fitsmode)
    except OSError:
        raise PhotoopException('Unable to read FLIST file.')
    #
    # Construct the scores filling in the values to FLIST.SCORE
    #
    flist[1].data['SCORE'][:] = sdss_score(flist)
    if rescore:
        flist.writeto(os.path.join(resolve_dir, 'window_flist_rescore.fits'))
    flist.close()
    #
    # Restore the PHOTO_CALIB variable
    #
    os.environ['PHOTO_CALIB'] = calib_dir_save
    return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
import warnings
import json
import stat
import sys
from os import chmod, remove
from os.path import exists, join
from shutil import copy, rmtree
from tempfile import mkdtemp
from time import sleep
from collections import OrderedDict
import pytest
import numpy as np
from astropy.utils.data import get_pkg_data_filename
from astropy.table import Table
from astropy.io.registry import (register_identifier, register_reader,
                                 register_writer)
from .. import PydlutilsException, PydlutilsUserWarning
from ..yanny import (write_ndarray_to_yanny, yanny, is_yanny,
                     read_table_yanny, write_table_yanny)


register_identifier('yanny', Table, is_yanny)
register_reader('yanny', Table, read_table_yanny)
register_writer('yanny', Table, write_table_yanny)


class YannyTestCase(object):
    """Based on astropy.io.fits.tests.FitsTestCase.
    """
    save_temp = False

    def setup_method(self):
        self.temp_dir = mkdtemp(prefix='yanny-test-')
        # Ignore deprecation warnings--this only affects Python 2.5 and 2.6,
        # since deprecation warnings are ignored by defualt on 2.7
        # warnings.simplefilter('ignore')
        # warnings.simplefilter('always', UserWarning)
        # raise ValueError("I am setting up a subclass of YannyTestCase!")
        with open(get_pkg_data_filename("t/yanny_data.json")) as js:
            self.test_data = json.load(js)

    def teardown_method(self):
        warnings.resetwarnings()
        if not self.save_temp:
            if hasattr(self, 'temp_dir') and exists(self.temp_dir):
                tries = 3
                while tries:
                    try:
                        rmtree(self.temp_dir)
                        break
                    except OSError:
                        # Probably couldn't delete the file because for
                        # whatever reason a handle to it is still open/hasn't
                        # been garbage-collected.
                        sleep(0.5)
                        tries -= 1
        # raise ValueError("I am tearing down up a subclass of YannyTestCase!")

    def copy_file(self, filename):
        """Copies a backup of a test data file to the temp dir and sets its
        mode to writeable.
        """
        copy(self.data(filename), self.temp(filename))
        chmod(self.temp(filename), stat.S_IREAD | stat.S_IWRITE)

    def data(self, filename):
        """Returns the path to a test data file.
        """
        return get_pkg_data_filename('t/'+filename)

    def temp(self, filename):
        """Returns the full path to a file in the test temp dir.
        """
        return join(self.temp_dir, filename)

    def json2dtype(self, data):
        """Convert JSON-encoded dtype data into a real dtype object.
        """
        stuff = list()
        for k in data:
            if len(k) == 3:
                stuff.append(tuple([str(k[0]), str(k[1]), tuple(k[2])]))
            else:
                stuff.append(tuple([str(k[0]), str(k[1])]))
        d = np.dtype(stuff)
        if sys.byteorder == 'little':
            return d
        else:
            return d.newbyteorder()


class TestYanny(YannyTestCase):
    """Test class for yanny files.
    """
    save_temp = False

    def test_is_yanny(self):
        """Test the yanny identifier.
        """
        par = yanny(self.data('test.par'))
        assert is_yanny('read', None, None, par)
        assert is_yanny('read', 'test.par', None)
        with open(self.data('test.par'), 'rb') as fileobj:
            assert is_yanny('read', None, fileobj)

    def test_read_table_yanny(self):
        """Test reading to an astropy Table.
        """
        filename = self.data('test_table.par')
        with pytest.raises(PydlutilsException):
            t = Table.read(filename)
        with pytest.raises(KeyError):
            t = Table.read(filename, tablename='foo')
        t = Table.read(filename, tablename='test')
        assert isinstance(t.meta, OrderedDict)
        assert t.meta['name'] == 'first table'
        assert (t['a'] == np.array([1, 4, 5])).all()
        assert (t['c'] == np.array([b'x', b'y', b'z'])).all()

    def test_write_table_yanny(self):
        """Test writing an astropy Table to yanny.
        """
        filename = self.temp('table.par')
        a = [1, 4, 5]
        b = [2.0, 5.0, 8.2]
        c = [b'x', b'y', b'z']
        t = Table([a, b, c], names=('a', 'b', 'c'),
                  meta={'name': 'first table'},
                  dtype=('i8', 'f8', 'S1'))
        t.write(filename, tablename='test')
        par1 = yanny(filename)
        par2 = yanny(self.data('test_table.par'))
        assert par1 == par2
        remove(filename)

    def test_write_single_ndarray_to_yanny(self):
        """Test the write_ndarray_to_yanny function.
        """
        test_data = self.test_data['tempfile1.par']
        table = test_data['structures']['MYSTRUCT0']
        dt = self.json2dtype(table['dtype'])
        mystruct = np.zeros((table['size'],), dtype=dt)
        for col in table['dtype']:
            mystruct[col[0]] = np.array(table['data'][col[0]], dtype=col[1])
        enums = {'new_flag': test_data['enums']['new_flag']}
        par = write_ndarray_to_yanny(self.temp('tempfile1.par'),
                                     mystruct,
                                     structnames='magnitudes',
                                     enums=enums)
        assert (par._symbols['enum'][0] ==
                'typedef enum {\n    FALSE,\n    TRUE\n} BOOLEAN;')
        assert par._symbols['struct'][0] == 'typedef struct {\n    double ra;\n    double dec;\n    float mag[5];\n    int flags;\n    BOOLEAN new_flag;\n} MAGNITUDES;'
        for k, f in enumerate(('FALSE', 'TRUE', 'TRUE', 'FALSE')):
            assert par['MAGNITUDES']['new_flag'][k].decode() == f

    def test_write_multiple_ndarray_to_yanny(self):
        """Test the write_ndarray_to_yanny function.
        """
        test_data = self.test_data['tempfile1.par']
        table = test_data['structures']['MYSTRUCT0']
        dt = self.json2dtype(table['dtype'])
        mystruct = np.zeros((table['size'],), dtype=dt)
        for col in table['dtype']:
            mystruct[col[0]] = np.array(table['data'][col[0]], dtype=col[1])
        enums = test_data['enums']
        table = test_data['structures']['MY_STATUS']
        dt = self.json2dtype(table['dtype'])
        status = np.zeros((table['size'],), dtype=dt)
        for col in table['dtype']:
            status[col[0]] = np.array(table['data'][col[0]], dtype=col[1])
        enums = test_data['enums']
        par = write_ndarray_to_yanny(self.temp('tempfile2.par'),
                                     (mystruct, status),
                                     structnames=('magnitudes', 'my_status'),
                                     enums=enums,
                                     comments=['This is a test',
                                               'This is another test'])
        assert ('typedef enum {\n    FALSE,\n    TRUE\n} BOOLEAN;' in
                par._symbols['enum'])
        assert 'typedef enum {\n    FAILURE,\n    INCOMPLETE,\n    SUCCESS\n} STATUS;' in par._symbols['enum']
        assert 'typedef struct {\n    double ra;\n    double dec;\n    float mag[5];\n    int flags;\n    BOOLEAN new_flag;\n} MAGNITUDES;' in par._symbols['struct']
        assert 'typedef struct {\n    long timestamp;\n    STATUS state;\n} MY_STATUS;' in par._symbols['struct']
        for k, f in enumerate(('FALSE', 'TRUE', 'TRUE', 'FALSE')):
            assert par['MAGNITUDES']['new_flag'][k].decode() == f
        for k, f in enumerate(["SUCCESS", "SUCCESS", "FAILURE", "INCOMPLETE"]):
            assert par['MY_STATUS']['state'][k].decode() == f

    def test_write_ndarray_to_yanny_exceptions(self):
        """Make sure certain execptions are raised by the
        write_ndarray_to_yanny function.
        """
        test_data = self.test_data['tempfile1.par']
        table = test_data['structures']['MYSTRUCT0']
        dt = self.json2dtype(table['dtype'])
        mystruct = np.zeros((table['size'],), dtype=dt)
        for col in table['dtype']:
            mystruct[col[0]] = np.array(table['data'][col[0]], dtype=col[1])
        enums = {'new_flag': test_data['enums']['new_flag']}
        with pytest.raises(PydlutilsException):
            _ = write_ndarray_to_yanny(self.data('test.par'), mystruct,
                                       structnames='magnitudes', enums=enums)
        with pytest.raises(PydlutilsException):
            _ = write_ndarray_to_yanny(self.temp('tempfile3.par'), mystruct,
                                       structnames=('magnitudes', 'my_status'),
                                       enums=enums)

    def test_write_ndarray_to_yanny_misc(self):
        """Test other minor features of write_ndarray_to_yanny.
        """
        test_data = self.test_data["tempfile1.par"]
        table = test_data['structures']['MYSTRUCT0']
        dt = self.json2dtype(table['dtype'])
        mystruct = np.zeros((table['size'],), dtype=dt)
        for col in table['dtype']:
            mystruct[col[0]] = np.array(table['data'][col[0]], dtype=col[1])
        enums = {'new_flag': test_data['enums']['new_flag']}
        hdr = test_data['pairs']
        par = write_ndarray_to_yanny(self.temp('tempfile1.par'), mystruct,
                                     hdr=hdr, enums=enums)
        assert 'MYSTRUCT0' in par
        assert par['keyword1'] == hdr['keyword1']
        assert par['keyword2'] == hdr['keyword2']

    def test_yanny(self):
        """Used to test the yanny class.
        """
        #
        # Describe what should be in the object
        #
        pair_data = self.test_data["test.par"]["pairs"]
        struct_data = self.test_data["test.par"]["structures"]
        s1 = """typedef struct {
    float mag[5];
    char b[5][];
    char foo[25];
    double c;
    int flags[2];
    BOOLEAN new_flag;
} MYSTRUCT;"""
        s2 = """typedef struct {
    float foo<3>; # This is archaic array notation, strongly deprecated,
    char bar<10>; # but still technically supported.
} OLD;"""
        s3 = """typedef struct {
    STATUS state;
    char timestamp[]; #UTC timestamp in format 2008-06-21T00:27:33
} STATUS_UPDATE;"""
        symbols = [s1, s2, s3]
        e1 = """typedef enum {
    FALSE,
    TRUE
} BOOLEAN;"""
        e2 = """typedef enum {
    FAILURE,
    INCOMPLETE,
    SUCCESS
} STATUS;"""
        enum = [e1, e2]
        with open(self.data('test.par')) as f:
            file_data = f.read()
        #
        # Open the object
        #
        par = yanny(self.data('test.par'))
        #
        # Test some static methods, etc.
        #
        assert par.get_token('abcd') == ('abcd', '')
        assert str(par) == file_data
        assert repr(par) == file_data
        par2 = yanny(self.data('test_table.par'))
        assert par != par2
        #
        # Test types
        #
        assert par.type('MYSTRUCT', 'c') == 'double'
        assert par.type('FOOBAR', 'd') is None
        assert par.type('MYSTRUCT', 'foobar') is None
        assert not par2.isenum('TEST', 'a')
        assert par.array_length('MYSTRUCT', 'c') == 1
        assert par.char_length('MYSTRUCT', 'c') is None
        #
        # Test the pairs
        #
        assert set(par.pairs()) == set(pair_data)
        #
        # Test the pair values
        #
        for p in par.pairs():
            assert par[p] == pair_data[p]
        #
        # Test the structure of the object
        #
        assert (set(par) == set(list(pair_data) +
                list(struct_data)))
        assert (set(par._symbols) == set(list(struct_data) +
                ['struct', 'enum']))
        assert set(par._symbols['struct']) == set(symbols)
        assert set(par._symbols['enum']) == set(enum)
        assert set(par.tables()) == set(struct_data)
        for t in par.tables():
            assert (par.dtype(t) ==
                    self.json2dtype(struct_data[t]['dtype']))
            assert par.size(t) == struct_data[t]['size']
            assert (set(par.columns(t)) ==
                    set(struct_data[t]['columns']))
            for c in par.columns(t):
                assert (par.type(t, c) ==
                        struct_data[t]['columns'][c])
                # print(par[t][c])
        assert par.isenum('MYSTRUCT', 'new_flag')
        assert par._enum_cache['BOOLEAN'] == ['FALSE', 'TRUE']
        assert (par._enum_cache['STATUS'] ==
                ['FAILURE', 'INCOMPLETE', 'SUCCESS'])
        #
        # Test values
        #
        assert np.allclose(par['MYSTRUCT'].mag[0],
                           np.array([17.5, 17.546, 17.4, 16.1, 16.0]))
        assert np.allclose(par['MYSTRUCT'].mag[5],
                           np.array([19.3, 18.2, 17.1, 16.0, 15.9]))
        assert par['MYSTRUCT'].foo[1] == b"My dog has no nose."
        assert np.allclose(par['MYSTRUCT'].c[2], 7.24345567)
        assert (par['MYSTRUCT']['flags'][2] == np.array([123123, 0])).all()
        #
        # Test expected write failures.
        #
        # This should fail, since test.par already exists.
        with pytest.raises(PydlutilsException):
            par.write()
        with pytest.warns(PydlutilsUserWarning):
            par.append({})
        datatable = {'status_update': {'state': ['SUCCESS', 'SUCCESS'],
                                       'timestamp': ['2008-06-22 01:27:33', '2008-06-22 01:27:36']},
                     'new_keyword': 'new_value'}
        par.filename = self.temp('test_append.par')
        # This should also fail, because test_append.par does not exist.
        with pytest.raises(PydlutilsException):
            par.append(datatable)
        return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-


def file_lines(path, compress=False):
    """Replicates the IDL ``FILE_LINES()`` function.

    Given a path to a file name or a list of such paths, returns the number of
    lines in the file(s).

    Parameters
    ----------
    path : :class:`str` or :class:`list` of :class:`str`
        Path to a file.  Can be a list of paths.
    compress : :class:`bool`, optional
        If set to ``True``, assumes that all files in `path` are GZIP
        compressed.

    Returns
    -------
    :class:`int` or :class:`list` of :class:`int`
        The number of lines in `path`.  Returns a list of lengths if a list of
        files is supplied.

    Notes
    -----
    The ``/NOEXPAND_PATH`` option in IDL's ``FILE_LINES()`` is not implemented.

    References
    ----------
    https://www.nv5geospatialsoftware.com/docs/file_lines.html

    Examples
    --------
    >>> from pydl import file_lines
    >>> from os.path import dirname, join
    >>> file_lines(join(dirname(__file__),'tests','t','this-file-contains-42-lines.txt'))
    42
    """
    scalar = False
    if isinstance(path, (str,)):
        working_path = [path]
        scalar = True
    else:
        working_path = path
    lines = list()
    for filename in working_path:
        if compress:
            #
            # gzip in Python 2.6 can't use a context manager.
            #
            import gzip
            f = gzip.open(filename)
            lines.append(len(f.readlines()))
            f.close()
        else:
            with open(filename) as f:
                lines.append(len(f.readlines()))
    if scalar:
        return lines[0]
    else:
        return lines



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the bspline directory in idlutils.
"""
from warnings import warn
import numpy as np
from numpy.linalg.linalg import LinAlgError
from scipy.linalg import cholesky_banded, cho_solve_banded
from . import PydlutilsUserWarning
from .math import djs_reject
from .trace import fchebyshev
from .. import uniq
from ..goddard.math import flegendre


class bspline(object):
    """B-spline class.

    Functions in the idlutils bspline library are implemented as methods on this
    class.

    Parameters
    ----------
    x : :class:`numpy.ndarray`
        The data.
    nord : :class:`int`, optional
        The order of the B-spline.  Default is 4, which is cubic.
    npoly : :class:`int`, optional
        Polynomial order to fit over 2nd variable, if supplied.  If not
        supplied the order is 1.
    bkpt : :class:`numpy.ndarray`, optional
        An initial breakpoint vector. If omitted, breakpoints will be
        computed from other options.
    bkspread : :class:`float`, optional
        Applies a scale factor to the difference between successive value of `bkpt`.
    placed : :class:`numpy.ndarray`, optional
        Precalculated breakpoint positions.
    bkspace : :class:`float`, optional
        Spacing of breakpoints in units of `x`.
    nbkpts : :class:`int`, optional
        Number of breakpoints to span `x` range; minimum is 2 (the endpoints).
    everyn : :class:`int`, optional
        Spacing of breakpoints in good pixels.

    Attributes
    ----------
    breakpoints
        The processed breakpoints for the fit, based on the initialization options.
    nord
        The order of the B-spline.  Default is 4, which is cubic.
    npoly
        Polynomial order to fit over 2nd variable, if supplied.  If not
        supplied the order is 1.
    mask
        A mask for `breakpoints`.
    coeff
        Internal attribute used by :meth:`~pydl.pydlutils.bspline.bspline.fit`.
    icoeff
        Internal attribute used by :meth:`~pydl.pydlutils.bspline.bspline.fit`.
    xmin
        Normalization minimum for the second variable in 2-dimensional fits.
    xmax
        Normalization maximum for the second variable in 2-dimensional fits.
    funcname
        For 2-dimensional fits, this is the function for the second variable.
        The default is 'legendre'.

    Notes
    -----
    Although this code has been tested (in the sense that it reproduces the
    original idlutils code) and appears to work for a subset 1-dimensional cases,
    it should not be used for 2 dimensions or more exotic 1-dimensional cases.
    """

    def __init__(self, x, nord=4, npoly=1, bkpt=None, bkspread=1.0,
                 placed=None, bkspace=None, nbkpts=None, everyn=None):
        """Init creates an object whose attributes are similar to the
        structure returned by the ``create_bsplineset()`` function.
        """
        #
        # Set the breakpoints.
        #
        if bkpt is None:
            startx = x.min()
            rangex = x.max() - startx
            if placed is not None:
                w = ((placed >= startx) &
                     (placed <= startx+rangex))
                if w.sum() < 2:
                    bkpt = np.arange(2, dtype='f') * rangex + startx
                else:
                    bkpt = placed[w]
            elif bkspace is not None:
                nbkpts = int(rangex/bkspace) + 1
                if nbkpts < 2:
                    nbkpts = 2
                tempbkspace = rangex/float(nbkpts-1)
                bkpt = np.arange(nbkpts, dtype='f')*tempbkspace + startx
            elif nbkpts is not None:
                if nbkpts < 2:
                    nbkpts = 2
                tempbkspace = rangex/float(nbkpts-1)
                bkpt = np.arange(nbkpts, dtype='f') * tempbkspace + startx
            elif everyn is not None:
                nx = x.size
                nbkpts = max(nx//everyn, 1)
                if nbkpts == 1:
                    xspot = [0]
                else:
                    xspot = int(nx/(nbkpts-1)) * np.arange(nbkpts, dtype='i4')
                bkpt = x[xspot].astype('f')
            else:
                raise ValueError('No information for bkpts.')
        imin = bkpt.argmin()
        imax = bkpt.argmax()
        if x.min() < bkpt[imin]:
            warn('Lowest breakpoint does not cover lowest x value: changing.',
                 PydlutilsUserWarning)
            bkpt[imin] = x.min()
        if x.max() > bkpt[imax]:
            warn('Highest breakpoint does not cover highest x value: changing.',
                 PydlutilsUserWarning)
            bkpt[imax] = x.max()
        nshortbkpt = bkpt.size
        fullbkpt = bkpt.copy()
        if nshortbkpt == 1:
            bkspace = np.float32(bkspread)
        else:
            bkspace = (bkpt[1] - bkpt[0]) * np.float32(bkspread)
        for i in np.arange(1, nord, dtype=np.float32):
            fullbkpt = np.insert(fullbkpt, 0, bkpt[0]-bkspace*i)
            fullbkpt = np.insert(fullbkpt, fullbkpt.shape[0],
                                 bkpt[nshortbkpt-1] + bkspace*i)
        #
        # Set the attributes
        #
        nc = fullbkpt.size - nord
        self.breakpoints = fullbkpt
        self.nord = nord
        self.npoly = npoly
        self.mask = np.ones((fullbkpt.size,), dtype='bool')
        if npoly > 1:
            self.coeff = np.zeros((npoly, nc), dtype='d')
            self.icoeff = np.zeros((npoly, nc), dtype='d')
        else:
            self.coeff = np.zeros((nc,), dtype='d')
            self.icoeff = np.zeros((nc,), dtype='d')
        self.xmin = 0.0
        self.xmax = 1.0
        self.funcname = 'legendre'
        return

    def fit(self, xdata, ydata, invvar, x2=None):
        """Calculate a B-spline in the least-squares sense.

        Fit is based on two variables: `xdata` which is sorted and spans a large range
        where breakpoints are required `ydata` which can be described with a low order
        polynomial.

        Parameters
        ----------
        xdata : :class:`numpy.ndarray`
            Independent variable.
        ydata : :class:`numpy.ndarray`
            Dependent variable.
        invvar : :class:`numpy.ndarray`
            Inverse variance of `ydata`.
        x2 : :class:`numpy.ndarray`, optional
            Orthogonal dependent variable for 2d fits.

        Returns
        -------
        :class:`tuple`
            A tuple containing an integer error code, and the evaluation of the
            b-spline at the input values.  An error code of -2 is a failure,
            -1 indicates dropped breakpoints, 0 is success, and positive
            integers indicate ill-conditioned breakpoints.
        """
        goodbk = self.mask[self.nord:]
        nn = goodbk.sum()
        if nn < self.nord:
            yfit = np.zeros(ydata.shape, dtype='f')
            return (-2, yfit)
        nfull = nn * self.npoly
        bw = self.npoly * self.nord
        a1, lower, upper = self.action(xdata, x2=x2)
        foo = np.tile(invvar, bw).reshape(bw, invvar.size).transpose()
        a2 = a1 * foo
        alpha = np.zeros((bw, nfull+bw), dtype='d')
        beta = np.zeros((nfull+bw,), dtype='d')
        bi = np.arange(bw, dtype='i4')
        bo = np.arange(bw, dtype='i4')
        for k in range(1, bw):
            bi = np.append(bi, np.arange(bw-k, dtype='i4')+(bw+1)*k)
            bo = np.append(bo, np.arange(bw-k, dtype='i4')+bw*k)
        for k in range(nn-self.nord+1):
            itop = k*self.npoly
            ibottom = min(itop, nfull) + bw - 1
            ict = upper[k] - lower[k] + 1
            if ict > 0:
                work = np.dot(a1[lower[k]:upper[k]+1, :].T, a2[lower[k]:upper[k]+1, :])
                wb = np.dot(ydata[lower[k]:upper[k]+1], a2[lower[k]:upper[k]+1, :])
                alpha.T.flat[bo+itop*bw] += work.flat[bi]
                beta[itop:ibottom+1] += wb
        min_influence = 1.0e-10 * invvar.sum() / nfull
        errb = cholesky_band(alpha, mininf=min_influence)
        if isinstance(errb[0], int) and errb[0] == -1:
            a = errb[1]
        else:
            yfit, foo = self.value(xdata, x2=x2, action=a1, upper=upper, lower=lower)
            return (self.maskpoints(errb[0]), yfit)
        sol = cholesky_solve(a, beta)
        if self.npoly > 1:
            self.icoeff[:, goodbk] = np.array(a[0, 0:nfull].reshape(self.npoly, nn), dtype=a.dtype)
            self.coeff[:, goodbk] = np.array(sol[0:nfull].reshape(self.npoly, nn), dtype=sol.dtype)
        else:
            self.icoeff[goodbk] = np.array(a[0, 0:nfull], dtype=a.dtype)
            self.coeff[goodbk] = np.array(sol[0:nfull], dtype=sol.dtype)
        yfit, foo = self.value(xdata, x2=x2, action=a1, upper=upper, lower=lower)
        return (0, yfit)

    def action(self, x, x2=None):
        """Construct banded B-spline matrix, with dimensions [ndata, bandwidth].

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            Independent variable.
        x2 : :class:`numpy.ndarray`, optional
            Orthogonal dependent variable for 2d fits.

        Returns
        -------
        :class:`tuple`
            A tuple containing the B-spline action matrix; the 'lower' parameter,
            a list of pixel positions, each corresponding to the first
            occurence of position greater than breakpoint indx; and 'upper',
            Same as lower, but denotes the upper pixel positions.
        """
        nx = x.size
        nbkpt = self.mask.sum()
        if nbkpt < 2*self.nord:
            return (-2, 0, 0)
        n = nbkpt - self.nord
        # gb = self.breakpoints[self.mask]
        bw = self.npoly*self.nord
        lower = np.zeros((n - self.nord + 1,), dtype='i4')
        upper = np.zeros((n - self.nord + 1,), dtype='i4') - 1
        indx = self.intrv(x)
        bf1 = self.bsplvn(x, indx)
        action = bf1
        aa = uniq(indx, np.arange(indx.size, dtype='i4'))
        upper[indx[aa]-self.nord+1] = aa
        rindx = indx[::-1]
        bb = uniq(rindx, np.arange(rindx.size, dtype='i4'))
        lower[rindx[bb]-self.nord+1] = nx - bb - 1
        if x2 is not None:
            if x2.size != nx:
                raise ValueError('Dimensions of x and x2 do not match.')
            x2norm = 2.0 * (x2 - self.xmin) / (self.xmax - self.xmin) - 1.0
            if self.funcname == 'poly':
                temppoly = np.ones((nx, self.npoly), dtype='f')
                for i in range(1, self.npoly):
                    temppoly[:, i] = temppoly[:, i-1] * x2norm
            elif self.funcname == 'poly1':
                temppoly = np.tile(x2norm, self.npoly).reshape(nx, self.npoly)
                for i in range(1, self.npoly):
                    temppoly[:, i] = temppoly[:, i-1] * x2norm
            elif self.funcname == 'chebyshev':
                temppoly = fchebyshev(x2norm, self.npoly)
            elif self.funcname == 'legendre':
                temppoly = flegendre(x2norm, self.npoly)
            else:
                raise ValueError('Unknown value of funcname.')
            action = np.zeros((nx, bw), dtype='d')
            counter = -1
            for ii in range(self.nord):
                for jj in range(self.npoly):
                    counter += 1
                    action[:, counter] = bf1[:, ii]*temppoly[:, jj]
        return (action, lower, upper)

    def intrv(self, x):
        """Find the segment between breakpoints which contain each value in the array `x`.

        The minimum breakpoint is ``nbkptord - 1``, and the maximum
        is ``nbkpt - nbkptord - 1``.

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            Data values, assumed to be monotonically increasing.

        Returns
        -------
        :class:`numpy.ndarray`
            Position of array elements with respect to breakpoints.
        """
        gb = self.breakpoints[self.mask]
        n = gb.size - self.nord
        indx = np.zeros((x.size,), dtype='i4')
        ileft = self.nord - 1
        for i in range(x.size):
            while x[i] > gb[ileft+1] and ileft < n - 1:
                ileft += 1
            indx[i] = ileft
        return indx

    def bsplvn(self, x, ileft):
        """Calculates the value of all possibly nonzero B-splines at `x`
        of a certain order.

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            Independent variable.
        ileft : :class:`int`
            Breakpoint segements that contain `x`.

        Returns
        -------
        :class:`numpy.ndarray`
            B-spline values.
        """
        bkpt = self.breakpoints[self.mask]
        vnikx = np.zeros((x.size, self.nord), dtype=x.dtype)
        deltap = vnikx.copy()
        deltam = vnikx.copy()
        j = 0
        vnikx[:, 0] = 1.0
        while j < self.nord - 1:
            ipj = ileft+j+1
            deltap[:, j] = bkpt[ipj] - x
            imj = ileft-j
            deltam[:, j] = x - bkpt[imj]
            vmprev = 0.0
            for l in range(j+1):
                vm = vnikx[:, l]/(deltap[:, l] + deltam[:, j-l])
                vnikx[:, l] = vm*deltap[:, l] + vmprev
                vmprev = vm*deltam[:, j-l]
            j += 1
            vnikx[:, j] = vmprev
        return vnikx

    def value(self, x, x2=None, action=None, lower=None, upper=None):
        """Evaluate a B-spline at specified values.

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            Independent variable.
        x2 : :class:`numpy.ndarray`, optional
            Orthogonal dependent variable for 2d fits.
        action : :class:`numpy.ndarray`, optional
            Action matrix to use.  If not supplied it is calculated.
        lower : :class:`numpy.ndarray`, optional
            If the action parameter is supplied, this parameter must also
            be supplied.
        upper : :class:`numpy.ndarray`, optional
            If the action parameter is supplied, this parameter must also
            be supplied.

        Returns
        -------
        :class:`tuple`
            A tuple containing the results of the bspline evaluation and a
            mask indicating where the evaluation was good.
        """
        xsort = x.argsort()
        xwork = x[xsort]
        if x2 is not None:
            x2work = x2[xsort]
        else:
            x2work = None
        if action is not None:
            if lower is None or upper is None:
                raise ValueError('Must specify lower and upper if action is set.')
        else:
            action, lower, upper = self.action(xwork, x2=x2work)
        yfit = np.zeros(x.shape, dtype=x.dtype)
        bw = self.npoly * self.nord
        spot = np.arange(bw, dtype='i4')
        goodbk = self.mask.nonzero()[0]
        coeffbk = self.mask[self.nord:].nonzero()[0]
        n = self.mask.sum() - self.nord
        if self.npoly > 1:
            goodcoeff = self.coeff[:, coeffbk]
        else:
            goodcoeff = self.coeff[coeffbk]
        # maskthis = np.zeros(xwork.shape,dtype=xwork.dtype)
        for i in range(n-self.nord+1):
            ict = upper[i] - lower[i] + 1
            if ict > 0:
                yfit[lower[i]:upper[i]+1] = np.dot(
                    action[lower[i]:upper[i]+1, :], goodcoeff[i*self.npoly+spot])
        yy = yfit.copy()
        yy[xsort] = yfit
        mask = np.ones(x.shape, dtype='bool')
        gb = self.breakpoints[goodbk]
        outside = ((x < gb[self.nord-1]) | (x > gb[n]))
        if outside.any():
            mask[outside] = False
        hmm = ((np.diff(goodbk) > 2).nonzero())[0]
        for jj in range(hmm.size):
            inside = ((x >= self.breakpoints[goodbk[hmm[jj]]]) &
                      (x <= self.breakpoints[goodbk[hmm[jj]+1]-1]))
            if inside.any():
                mask[inside] = False
        return (yy, mask)

    def maskpoints(self, err):
        """Perform simple logic of which breakpoints to mask.

        Parameters
        ----------
        err : :class:`numpy.ndarray`
            The list of indexes returned by the cholesky routines.

        Returns
        -------
        :class:`int`
            An integer indicating the results of the masking.  -1 indicates
            that the error points were successfully masked.  -2 indicates
            failure; the calculation should be aborted.

        Notes
        -----
        The mask attribute is modified, assuming it is possible to create the
        mask.
        """
        nbkpt = self.mask.sum()
        if nbkpt <= 2*self.nord:
            return -2
        hmm = err[np.unique(err/self.npoly)]/self.npoly
        n = nbkpt - self.nord
        if np.any(hmm >= n):
            return -2
        test = np.zeros(nbkpt, dtype='bool')
        for jj in range(-np.ceil(self.nord/2.0), self.nord/2.0):
            foo = np.where((hmm+jj) > 0, hmm+jj, np.zeros(hmm.shape, dtype=hmm.dtype))
            inside = np.where((foo+self.nord) < n-1, foo+self.nord, np.zeros(hmm.shape, dtype=hmm.dtype)+n-1)
            test[inside] = True
        if test.any():
            reality = self.mask[test]
            if self.mask[reality].any():
                self.mask[reality] = False
                return -1
            else:
                return -2
        else:
            return -2


def cholesky_band(l, mininf=0.0):
    """Compute *lower* Cholesky decomposition of a banded matrix.

    This function provides informative error messages to pass back to the
    :class:`~pydl.pydlutils.bspline.bspline` machinery; the actual
    computation is delegated to :func:`scipy.linalg.cholesky_banded`.

    Parameters
    ----------
    l : :class:`numpy.ndarray`
        A matrix on which to perform the Cholesky decomposition.  The
        matrix must be in a special, *lower* form described in
        :func:`scipy.linalg.cholesky_banded`.  In addition, the input
        must be padded.  If the original, square matrix has size
        :math:`N \\times N`, and the width of the band is :math:`b`,
        `l` must be :math:`b \\times (N + b)`.
    mininf : :class:`float`, optional
        Entries in the `l` matrix are considered negative if they are less
        than this value (default 0.0).

    Returns
    -------
    :class:`tuple`
        If problems were detected, the first item will be the index or
        indexes where the problem was detected, and the second item will simply
        be the input matrix.  If no problems were detected, the first item
        will be -1, and the second item will be the Cholesky decomposition.
    """
    bw, nn = l.shape
    n = nn - bw
    negative = l[0, 0:n] <= mininf
    if negative.any() or not np.all(np.isfinite(l)):
        warn('Bad entries: ' + str(negative.nonzero()[0]), PydlutilsUserWarning)
        return (negative.nonzero()[0], l)
    try:
        lower = cholesky_banded(l[:, 0:n], lower=True)
    except LinAlgError:
        #
        # Figure out where the error is.
        #
        lower = l.copy()
        kn = bw - 1
        spot = np.arange(kn, dtype='i4') + 1
        for j in range(n):
            lower[0, j] = np.sqrt(lower[0, j])
            lower[spot, j] /= lower[0, j]
            x = lower[spot, j]
            if not np.all(np.isfinite(x)):
                warn('NaN found in cholesky_band.', PydlutilsUserWarning)
                return (j, l)
    #
    # Restore padding.
    #
    L = np.zeros(l.shape, dtype=l.dtype)
    L[:, 0:n] = lower
    return (-1, L)


def cholesky_solve(a, bb):
    """Solve the equation :math:`A x = b` where `a` is a *lower*
    Cholesky-banded matrix.

    In the :class:`~pydl.pydlutils.bspline.bspline` machinery, `a` needs to
    be padded.  This function should only used with the output of
    :func:`~pydl.pydlutils.bspline.cholesky_band`, to ensure the proper
    padding on `a`.  Otherwise the computation is delegated to
    :func:`scipy.linalg.cho_solve_banded`.

    Parameters
    ----------
    a : :class:`numpy.ndarray`
        *Lower* Cholesky decomposition of :math:`A` in :math:`A x = b`.
    bb : :class:`numpy.ndarray`
        :math:`b` in :math:`A x = b`.

    Returns
    -------
    :class:`numpy.ndarray`
        The solution, padded to be the same shape as `bb`.
    """
    bw = a.shape[0]
    n = bb.shape[0] - bw
    x = np.zeros(bb.shape, dtype=bb.dtype)
    x[0:n] = cho_solve_banded((a[:, 0:n], True), bb[0:n])
    return x


def iterfit(xdata, ydata, invvar=None, upper=5, lower=5, x2=None,
            maxiter=10, **kwargs):
    """Iteratively fit a B-spline set to data, with rejection.

    Additional keyword parameters are passed to
    :class:`~pydl.pydlutils.bspline.bspline`.

    Parameters
    ----------
    xdata : :class:`numpy.ndarray`
        Independent variable.
    ydata : :class:`numpy.ndarray`
        Dependent variable.
    invvar : :class:`numpy.ndarray`, optional
        Inverse variance of `ydata`.  If not set, it will be calculated based
        on the standard deviation.
    upper : :class:`int` or :class:`float`, optional
        Upper rejection threshold in units of sigma, defaults to 5 sigma.
    lower : :class:`int` or :class:`float`, optional
        Lower rejection threshold in units of sigma, defaults to 5 sigma.
    x2 : :class:`numpy.ndarray`, optional
        Orthogonal dependent variable for 2d fits.
    maxiter : :class:`int`, optional
        Maximum number of rejection iterations, default 10.  Set this to
        zero to disable rejection.

    Returns
    -------
    :class:`tuple`
        A tuple containing the fitted bspline object and an output mask.
    """
    nx = xdata.size
    if ydata.size != nx:
        raise ValueError('Dimensions of xdata and ydata do not agree.')
    if invvar is not None:
        if invvar.size != nx:
            raise ValueError('Dimensions of xdata and invvar do not agree.')
    else:
        #
        # This correction to the variance makes it the same
        # as IDL's variance()
        #
        var = ydata.var()*(float(nx)/float(nx-1))
        if var == 0:
            var = 1.0
        invvar = np.ones(ydata.shape, dtype=ydata.dtype)/var
    if x2 is not None:
        if x2.size != nx:
            raise ValueError('Dimensions of xdata and x2 do not agree.')
    yfit = np.zeros(ydata.shape, dtype=ydata.dtype)
    if invvar.size == 1:
        outmask = True
    else:
        outmask = np.ones(invvar.shape, dtype='bool')
    xsort = xdata.argsort()
    maskwork = (outmask & (invvar > 0))[xsort]
    if 'oldset' in kwargs:
        sset = kwargs['oldset']
        sset.mask = True
        sset.coeff = 0
    else:
        if not maskwork.any():
            raise ValueError('No valid data points.')
        if 'fullbkpt' in kwargs:
            # fullbkpt = kwargs['fullbkpt']
            raise ValueError('Input via fullbkpt is not supported!')
        else:
            sset = bspline(xdata[xsort[maskwork]], **kwargs)
            if maskwork.sum() < sset.nord:
                warn('Number of good data points fewer than nord.',
                     PydlutilsUserWarning)
                return (sset, outmask)
            if x2 is not None:
                if 'xmin' in kwargs:
                    xmin = kwargs['xmin']
                else:
                    xmin = x2.min()
                if 'xmax' in kwargs:
                    xmax = kwargs['xmax']
                else:
                    xmax = x2.max()
                if xmin == xmax:
                    xmax = xmin + 1
                sset.xmin = xmin
                sset.xmax = xmax
                if 'funcname' in kwargs:
                    sset.funcname = kwargs['funcname']
    xwork = xdata[xsort]
    ywork = ydata[xsort]
    invwork = invvar[xsort]
    if x2 is not None:
        warn('2D bspline fits may be buggy and will be fully removed in the future.',
             DeprecationWarning)
        x2work = x2[xsort]
    else:
        x2work = None
    iiter = 0
    error = 0
    qdone = -1
    while (error != 0 or qdone == -1) and iiter <= maxiter:
        goodbk = sset.mask.nonzero()[0]
        if maskwork.sum() <= 1 or not sset.mask.any():
            sset.coeff = 0
            iiter = maxiter + 1
        else:
            if 'requiren' in kwargs:
                i = 0
                while xwork[i] < sset.breakpoints[goodbk[sset.nord]] and i < nx-1:
                    i += 1
                ct = 0
                for ileft in range(sset.nord, sset.mask.sum()-sset.nord+1):
                    while (xwork[i] >= sset.breakpoints[goodbk[ileft]] and
                           xwork[i] < sset.breakpoints[goodbk[ileft+1]] and
                           i < nx-1):
                        ct += invwork[i]*maskwork[i] > 0
                        i += 1
                    if ct >= kwargs['requiren']:
                        ct = 0
                    else:
                        sset.mask[goodbk[ileft]] = False
            error, yfit = sset.fit(xwork, ywork, invwork*maskwork,
                                   x2=x2work)
        iiter += 1
        inmask = maskwork
        if error == -2:
            return (sset, outmask)
        elif error == 0:
            maskwork, qdone = djs_reject(ywork, yfit, invvar=invwork,
                                         inmask=inmask, outmask=maskwork,
                                         upper=upper, lower=lower)
        else:
            pass
    outmask[xsort] = maskwork
    temp = yfit
    yfit[xsort] = temp
    return (sset, outmask)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
import pytest
import numpy as np
import pydl.pydlutils.sdss
from astropy.utils.data import get_pkg_data_filename
from ..sdss import (default_skyversion, sdss_flagexist, sdss_flagname,
                    sdss_flagval, set_maskbits, sdss_astrombad,
                    sdss_objid, sdss_specobjid, unwrap_specobjid)


class TestSDSS(object):
    """Test the functions in pydl.pydlutils.sdss.
    """

    def setup_method(self):
        pydl.pydlutils.sdss.maskbits = set_maskbits(
            maskbits_file=get_pkg_data_filename('t/testMaskbits.par'))
        self.opbadfields = np.array([(77, 'astrom', 30, 73, 'Large astrometric offset at field 39... 72'),
                                     (85, 'astrom', 8, 28, 'Large astrometric offset at field 11... 27'),
                                     (85, 'rotator', 242, 253, 'Large rotator offset at field 251 252'),
                                     (209, 'astrom', 8, 116, 'Tel. offsets before r-band field 115 -DJS'),
                                     (209, 'astrom', 137, 175, 'Tel. offsets after r-band field 145 -DJS'),
                                     (250, 'astrom', 456, 468, 'Large astrometric offset -Manual'),
                                     (251, 'astrom', 147, 159, 'Large astrometric offset at field 156 158')],
                                    dtype=[('run', '<i4'),
                                           ('problem', 'S8'),
                                           ('firstfield', '<i4'),
                                           ('lastfield', '<i4'),
                                           ('comments', 'S47')])
        pydl.pydlutils.sdss.opbadfields = self.opbadfields
        return

    def test_default_skyversion(self):
        assert default_skyversion() == 2

    def test_sdss_flagexist(self):
        assert sdss_flagexist('TARGET', 'ROSAT_A')
        assert sdss_flagexist('TARGET',
                              ['rosat_a', 'rosat_b', 'rosat_c', 'rosat_d'])
        l, f = sdss_flagexist('target', 'FOOBAR', flagexist=True)
        assert not l
        assert f
        l, which = sdss_flagexist('TARGET', ['rosat_a', 'rosat_b', 'rosat_c',
                                             'rosat_d', 'foobar'], whichexist=True)
        assert not l
        assert tuple(which) == (True, True, True, True, False)
        l, f, which = sdss_flagexist('TARGET', ['rosat_a', 'rosat_b',
                                                'rosat_c', 'rosat_d', 'foobar'],
                                     flagexist=True, whichexist=True)
        assert not l
        assert f
        assert tuple(which) == (True, True, True, True, False)

    def test_sdss_flagname(self):
        names = sdss_flagname('ANCILLARY_TARGET1', 2310346608843161600)
        assert tuple(names) == ('BRIGHTGAL', 'BLAZGX', 'ELG')
        names = sdss_flagname('ANCILLARY_TARGET1', 2310346608843161600,
                              concat=True)
        assert names == 'BRIGHTGAL BLAZGX ELG'
        with pytest.raises(KeyError):
            names = sdss_flagname('ABADMASK', 123456789)

    def test_sdss_flagval(self):
        val = sdss_flagval('TARGET', 'ROSAT_A')
        assert val == 2**9
        val = sdss_flagval('ANCILLARY_TARGET1', ['BLAZGX', 'ELG', 'BRIGHTGAL'])
        assert val == 2310346608843161600
        with pytest.raises(KeyError):
            val = sdss_flagval('TARGET', 'ROSAT_Q')
        with pytest.raises(KeyError):
            val = sdss_flagval('ABADMASK', "ABADFLAG")

    def test_set_maskbits(self):
        maskbits = pydl.pydlutils.sdss.maskbits
        assert (set(maskbits.keys()) ==
                set(['TARGET', 'BOSS_TARGET1', 'PRIMTARGET', 'ANCILLARY_TARGET1',
                     'ZWARNING', 'TTARGET', 'SECTARGET', 'LEGACY_TARGET2',
                     'LEGACY_TARGET1', 'SPECIAL_TARGET2', 'FLUXMATCH_STATUS']))
        assert (set(maskbits['TARGET'].keys()) ==
                set(['QSO_FIRST_SKIRT', 'QSO_CAP', 'GALAXY_RED', 'STAR_CARBON',
                     'STAR_WHITE_DWARF', 'GALAXY_RED_II', 'GALAXY_BIG',
                     'GALAXY_BRIGHT_CORE', 'SERENDIP_MANUAL', 'STAR_SUB_DWARF',
                     'QSO_FIRST_CAP', 'QSO_SKIRT', 'STAR_PN', 'STAR_BHB', 'QSO_HIZ',
                     'STAR_BROWN_DWARF', 'SERENDIP_FIRST', 'SOUTHERN_SURVEY',
                     'STAR_RED_DWARF', 'STAR_CATY_VAR', 'QSO_REJECT', 'GALAXY',
                     'SERENDIP_RED', 'SERENDIP_DISTANT', 'QSO_MAG_OUTLIER',
                     'ROSAT_A', 'ROSAT_C', 'ROSAT_B', 'ROSAT_E', 'ROSAT_D',
                     'SERENDIP_BLUE']))

    def test_sdss_astrombad(self):
        assert not sdss_astrombad(77, 1, 20)
        assert sdss_astrombad(77, 3, 35)
        assert not sdss_astrombad(77, 6, 77)
        assert sdss_astrombad(85, 1, 15)
        assert (sdss_astrombad(np.array([77, 85, 251]),
                np.array([1, 2, 3]), np.array([20, 15, 151])) ==
                np.array([False, True, True])).all()

    def test_sdss_astrombad_raises(self):
        with pytest.raises(ValueError):
            _ = sdss_astrombad(77, 32, 20)
        with pytest.raises(ValueError):
            _ = sdss_astrombad(-1, 1, 20)
        with pytest.raises(ValueError):
            _ = sdss_astrombad(2**17, 1, 20)
        with pytest.raises(ValueError):
            _ = sdss_astrombad(-2, 1, 20)
        with pytest.raises(ValueError):
            _ = sdss_astrombad(251, 1, 2**16)
        with pytest.raises(ValueError):
            _ = sdss_astrombad(np.array([77, 85, 251]), np.array([1]),
                               np.array([20, 15, 151]))
        with pytest.raises(ValueError):
            _ = sdss_astrombad(np.array([77, 85, 251]), np.array([1, 2, 3]),
                               np.array([20]))

    @pytest.mark.remote_data
    def test_sdss_astrombad_remote(self):
        pydl.pydlutils.sdss.opbadfields = None
        assert not sdss_astrombad(77, 1, 20)
        assert sdss_astrombad(77, 3, 35)
        assert not sdss_astrombad(77, 6, 77)

    def test_sdss_objid(self):
        assert sdss_objid(3704, 3, 91, 146) == 1237661382772195474
        run = np.array([3704, 1000], dtype=np.int64)
        camcol = np.array([3, 6], dtype=np.int64)
        field = np.array([91, 77], dtype=np.int64)
        obj = np.array([146, 123], dtype=np.int64)
        assert (np.array([1237661382772195474, 1237649770790322299],
                         dtype=np.int64) ==
                sdss_objid(run, camcol, field, obj)).all()
        assert (sdss_objid(752, 5, 618, 459, rerun=40,
                           skyversion=1, firstfield=1) == 587722984180548043)
        #
        # Exceptions
        #
        with pytest.raises(ValueError):
            _ = sdss_objid(run, 3, 91, 146)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, camcol, 91, 146)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, field, 146)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, obj)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, 146, rerun=np.array([137, 301]))
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, 146, skyversion=np.array([2, 3]))
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, 146, firstfield=np.array([0, 1]))
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, 146, skyversion=-2)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, 146, skyversion=16)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, 146, firstfield=-2)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, 146, firstfield=2)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, 146, rerun=-2)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, 146, rerun=2**11)
        with pytest.raises(ValueError):
            _ = sdss_objid(-2, 3, 91, 146)
        with pytest.raises(ValueError):
            _ = sdss_objid(2**16, 3, 91, 146)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 0, 91, 146)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 7, 91, 146)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, -2, 146)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 2**12, 146)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, -2)
        with pytest.raises(ValueError):
            _ = sdss_objid(3704, 3, 91, 2**16)

    def test_sdss_specobjid(self):
        s = sdss_specobjid(4055, 408, 55359, 'v5_7_0')
        assert s == 4565636362342690816
        s = sdss_specobjid(4055, 408, 55359, '700')
        assert s == 4565636362342690816
        s = sdss_specobjid(4055, 408, 55359, 700)
        assert s == 4565636362342690816
        s = sdss_specobjid(4055, 408, 55359, 'v5_7_0', line=137)
        assert s == 4565636362342690816 + 137
        s = sdss_specobjid(4055, 408, 55359, 'v5_7_0', index=137)
        assert s == 4565636362342690816 + 137
        #
        # Because 5-digit plate IDs...
        #
        s = sdss_specobjid(10000, 1, 57346, 'v5_10_0')
        assert s == 11258999466550599680
        #
        # Exceptions
        #
        with pytest.raises(ValueError):
            s = sdss_specobjid(np.array([4055, 4056]), 408, 55359, 'v5_7_0')
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, np.array([408, 409]), 55359, 'v5_7_0')
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, np.array([55359, 55360]), 'v5_7_0')
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 55359,
                               np.array(['v5_7_0', 'v5_7_2']))
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 55359, 'v5_7_0',
                               line=np.array([137, 138]))
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 55359, 'v5_7_0',
                               index=np.array([137, 138]))
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 55359, 'v5_7_0', line=137,
                               index=137)
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 55359, 'V5.7.0')
        with pytest.raises(ValueError):
            s = sdss_specobjid(-1, 408, 55359, 'v5_7_0')
        with pytest.raises(ValueError):
            s = sdss_specobjid(2**15, 408, 55359, 'v5_7_0')
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, -1, 55359, 'v5_7_0')
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 2**12, 55359, 'v5_7_0')
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 49999, 'v5_7_0')
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 2**15, 'v5_7_0')
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 55359, -1)
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 55359, 2**15)
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 55359, 'v5_7_0', line=-1)
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 55359, 'v5_7_0', line=2**10)
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 55359, 'v5_7_0', index=-1)
        with pytest.raises(ValueError):
            s = sdss_specobjid(4055, 408, 55359, 'v5_7_0', index=2**10)

    def test_unwrap_specobjid(self):
        s = 4565636362342690816
        d = unwrap_specobjid(np.array([s], dtype=np.uint64))
        assert d.plate == 4055
        assert d.mjd == 55359
        assert d.fiber == 408
        assert d.run2d == 'v5_7_0'
        assert d.line == 0
        d = unwrap_specobjid(np.array(['4565636362342690816']))
        assert d.plate == 4055
        assert d.mjd == 55359
        assert d.fiber == 408
        assert d.run2d == 'v5_7_0'
        assert d.line == 0
        si = 4565636362342690816 + 137
        d = unwrap_specobjid(np.array([si], dtype=np.uint64))
        assert d.plate == 4055
        assert d.mjd == 55359
        assert d.fiber == 408
        assert d.run2d == 'v5_7_0'
        assert d.line == 137
        d = unwrap_specobjid(np.array([si], dtype=np.uint64),
                             specLineIndex=True)
        assert d.plate == 4055
        assert d.mjd == 55359
        assert d.fiber == 408
        assert d.run2d == 'v5_7_0'
        assert d.index == 137
        d = unwrap_specobjid(np.array([si], dtype=np.uint64),
                             run2d_integer=True)
        assert d.plate == 4055
        assert d.mjd == 55359
        assert d.fiber == 408
        assert d.run2d == 700
        assert d.line == 137
        #
        # Because 5-digit plate IDs...
        #
        s5 = 11258999466550599680
        d = unwrap_specobjid(np.array([s5], dtype=np.uint64))
        assert d.plate == 10000
        assert d.mjd == 57346
        assert d.fiber == 1
        assert d.run2d == 'v5_10_0'
        assert d.line == 0
        #
        # Exceptions
        #
        with pytest.raises(ValueError):
            d = unwrap_specobjid(np.array([4565636362342690816],
                                 dtype=np.int64))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.photoop.sdssio.
"""
import os
import pytest
import numpy as np
from ..sdssio import (filtername, filternum, sdss_calib, sdss_name, sdss_path,
                      sdssflux2ab)


path_data = {
    'apObj': "/PHOTO_REDUX/301/137/objcs/4",
    'calibMatch': "/PHOTO_REDUX/301/137/nfcalib",
    'calibPhotom': "/PHOTO_REDUX/301/137/nfcalib",
    'calibPhotomGlobal': "/PHOTO_CALIB/301/137/nfcalib",
    'fakeIdR': "/PHOTO_DATA/137/fake_fields/4",
    'fpAtlas': "/PHOTO_REDUX/301/137/objcs/4",
    'fpBIN': "/PHOTO_REDUX/301/137/objcs/4",
    'fpC': "/PHOTO_REDUX/301/137/objcs/4",
    'fpFieldStat': "/PHOTO_REDUX/301/137/objcs/4",
    'fpM': "/PHOTO_REDUX/301/137/objcs/4",
    'fpObjc': "/PHOTO_REDUX/301/137/objcs/4",
    'hoggObj': "/PHOTO_REDUX/301/137/objcs/4",
    'idFF': "/PHOTO_REDUX/301/137/objcs/4",
    'idR': "/PHOTO_DATA/137/fields/4",
    'idRR': "/PHOTO_DATA/137/fields/4",
    'psBB': "/PHOTO_REDUX/301/137/objcs/4",
    'psFF': "/PHOTO_REDUX/301/137/objcs/4",
    'psField': "/PHOTO_REDUX/301/137/objcs/4",
    'reObjGlobal': "/PHOTO_RESOLVE/301/137/resolve/4",
    'reObjRun': "/PHOTO_REDUX/301/137/resolve/4",
    'reObjTmp': "/PHOTO_RESOLVE/301/137/resolve/4",
    'tsField': "/PHOTO_REDUX/301/137/calibChunks/4"}


name_data = {
    'apObj': "apObj-000137-r4-0042.fit",
    'calibMatch': "calibMatch-000137-4.fits",
    'calibPhotom': "calibPhotom-000137-4.fits",
    'calibPhotomGlobal': "calibPhotomGlobal-000137-4.fits",
    'fakeIdR': "idR-000137-r4-0042.fit",
    'fpAtlas': "fpAtlas-000137-4-0042.fit",
    'fpBIN': "fpBIN-000137-r4-0042.fit",
    'fpC': "fpC-000137-r4-0042.fit",
    'fpFieldStat': "fpFieldStat-000137-4-0042.fit",
    'fpM': "fpM-000137-r4-0042.fit",
    'fpObjc': "fpObjc-000137-4-0042.fit",
    'hoggObj': "hoggObj-000137-4-0042.fits",
    'idFF': "idFF-000137-r4.fit",
    'idR': "idR-000137-r4-0042.fit",
    'idRR': "idRR-000137-r4-0042.fit",
    'psBB': "psBB-000137-r4-0042.fit",
    'psFF': "psFF-000137-r4.fit",
    'psField': "psField-000137-4-0042.fit",
    'reObjGlobal': "reObjGlobal-000137-4-0042.fits",
    'reObjRun': "reObjRun-000137-4-0042.fits",
    'reObjTmp': "reObjTmp-000137-4-0042.fits",
    'tsField': "tsField-000137-4-301-0042.fit"}


@pytest.fixture
def sdss_env(request):
    """Set up environment variables for testing sdss_path and sdss_name.
    """
    m = request.getfixturevalue("monkeypatch")
    for p in ('PHOTO_CALIB', 'PHOTO_DATA', 'BOSS_PHOTOOBJ', 'PHOTO_REDUX',
              'PHOTO_RESOLVE', 'PHOTO_SKY', 'PHOTO_SWEEP'):
        m.setenv(p, '/' + p)
    return m


def test_filtername():
    assert filtername(0) == 'u'
    assert filtername(1) == 'g'
    assert filtername(2) == 'r'
    assert filtername(3) == 'i'
    assert filtername(4) == 'z'
    #
    # filtername should return its argument if it's not
    # integer-like
    #
    assert filtername('r') == 'r'


def test_filternum():
    assert filternum('u') == 0
    assert filternum('g') == 1
    assert filternum('r') == 2
    assert filternum('i') == 3
    assert filternum('z') == 4
    #
    # Test default return value
    #
    fn = filternum()
    for k in range(5):
        assert fn[k] == k


def test_sdss_calib():
    foo = sdss_calib(94, 6, 101)
    assert foo['NMGYPERCOUNT'] == 1.0


def test_sdss_name(sdss_env):
    #
    # Bad ftype
    #
    with pytest.raises(KeyError):
        _ = sdss_name('fooBar', 137, 4, 42)
    for ftype in name_data:
        assert sdss_name(ftype, 137, 4, 42, '301', 'r',
                         no_path=True) == name_data[ftype]


def test_sdss_name_with_path(sdss_env):
    for ftype in name_data:
        assert (sdss_name(ftype, 137, 4, 42, '301', 'r') ==
                os.path.join(path_data[ftype], name_data[ftype]))


def test_sdss_name_reObj(sdss_env):
    assert sdss_name('reObj', 137, 4, 42, '301', 'r',
                     no_path=True) == name_data['reObjGlobal']
    sdss_env.delenv('PHOTO_RESOLVE')
    assert sdss_name('reObj', 137, 4, 42, '301', 'r',
                     no_path=True) == name_data['reObjRun']


def test_sdss_path(sdss_env):
    with pytest.raises(KeyError):
        _ = sdss_path('fooBar', 137)
    for ftype in path_data:
        assert sdss_path(ftype, 137, 4, '301') == path_data[ftype]


def test_sdssflux2ab():
    correction = np.array([-0.042, 0.036, 0.015, 0.013, -0.002])
    mags = np.zeros((2, 5), dtype='d')
    mags[0, :] = 18.0
    mags[1, :] = 19.0
    ab = sdssflux2ab(mags, magnitude=True)
    assert (ab == (mags + correction)).all()
    flux = 10**((22.5 - mags)/2.5)  # nanomaggies
    ab = sdssflux2ab(flux)
    assert np.allclose(ab, 10**((22.5 - (mags + correction))/2.5))
    ivar = 1.0/flux
    ab = sdssflux2ab(ivar, ivar=True)
    assert np.allclose(ab, ivar/(10**(-2.0*correction/2.5)))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
import numpy as np
from astropy.utils import lazyproperty


class pcomp(object):
    """Replicates the IDL ``PCOMP()`` function.

    The attributes of this class are all read-only properties, implemented
    with :class:`~astropy.utils.decorators.lazyproperty`.

    Parameters
    ----------
    x : array-like
        A 2-D array with :math:`N` rows and :math:`M` columns.
    standardize : :class:`bool`, optional
        If set to ``True``, the input data will have its mean subtracted off
        and will be scaled to unit variance.
    covariance : :class:`bool`, optional.
        If set to ``True``, the covariance matrix of the data will be used for
        the computation.  Otherwise the correlation matrix will be used.

    Notes
    -----

    References
    ----------
    https://www.nv5geospatialsoftware.com/docs/pcomp.html

    Examples
    --------
    """

    def __init__(self, x, standardize=False, covariance=False):
        from scipy.linalg import eigh
        if x.ndim != 2:
            raise ValueError('Input array must be two-dimensional')
        no, nv = x.shape
        self._nv = nv
        if standardize:
            xstd = x - np.tile(x.mean(0), no).reshape(x.shape)
            s = np.tile(xstd.std(0), no).reshape(x.shape)
            self._array = xstd/s
            self._xstd = xstd
        else:
            self._array = x
            self._xstd = None
        self._standardize = standardize
        if covariance:
            self._c = np.cov(self._array, rowvar=0)
        else:
            self._c = np.corrcoef(self._array, rowvar=0)
        self._covariance = covariance
        #
        # eigh is used for symmetric matrices
        #
        evals, evecs = eigh(self._c)
        #
        # Sort eigenvalues in descending order
        #
        ie = evals.argsort()[::-1]
        self._evals = evals[ie]
        self._evecs = evecs[:, ie]
        #
        # If necessary, add code to fix the signs of the eigenvectors.
        # http://www3.interscience.wiley.com/journal/117912150/abstract
        #
        return

    @lazyproperty
    def coefficients(self):
        """(:class:`~numpy.ndarray`) The principal components.
        These are the coefficients of `derived`.
        Basically, they are a re-scaling of the eigenvectors.
        """
        return self._evecs * np.tile(np.sqrt(self._evals), self._nv).reshape(
            self._nv, self._nv)

    @lazyproperty
    def derived(self):
        """(:class:`~numpy.ndarray`) The derived variables.
        """
        derived_data = np.dot(self._array, self.coefficients)
        if self._standardize:
            derived_data += self._xstd
        return derived_data

    @lazyproperty
    def variance(self):
        """(:class:`~numpy.ndarray`) The variances of each derived variable.
        """
        return self._evals/self._c.trace()

    @lazyproperty
    def eigenvalues(self):
        """(:class:`~numpy.ndarray`) The eigenvalues.
        There is one eigenvalue for each principal component.
        """
        return self._evals



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.photoop.window.
"""
import os
import pytest
import numpy as np
from .. import PhotoopException
from ..window import window_read, window_score
from ...pydlutils.mangle import FITS_polygon


@pytest.fixture
def fits_open(request):
    data_size = 10
    m = request.getfixturevalue("mocker")
    hhh = m.MagicMock()
    hhh.data = {'SCORE': np.zeros((data_size,), dtype=np.int16),
                'ICAP': np.arange(data_size, dtype=np.int16),
                'NCAPS': np.ones((data_size,), dtype=np.int16),
                'X': np.zeros((data_size,), dtype=np.float32),
                'CM': np.zeros((data_size,), dtype=np.float32)}
    hh = m.MagicMock()
    hh.__enter__.return_value = [None, hhh]
    h = m.patch('astropy.io.fits.open')
    h.return_value = hh
    h.data_size = data_size
    return h


@pytest.fixture
def table_read(request):
    data_size = 10
    m = request.getfixturevalue("mocker")

    class fake_table(dict):
        def __len__(self):
            return data_size

    hh = fake_table()
    hh['SCORE'] = np.zeros((data_size,), dtype=np.int16)
    hh['IPRIMARY'] = np.zeros((data_size,), dtype=np.int16)
    hh['IBINDX'] = np.zeros((data_size,), dtype=np.int16)
    hh['ICAP'] = 4*np.arange(data_size, dtype=np.int16)
    hh['NCAPS'] = np.ones((data_size,), dtype=np.int16) + 3
    hh['WEIGHT'] = np.ones((data_size,), dtype=np.float32)
    hh['STR'] = np.zeros((data_size,), dtype=np.float32)
    hh['X'] = np.zeros((4*data_size, 3), dtype=np.float32)
    hh['CM'] = np.zeros((4*data_size,), dtype=np.float32)
    h = m.patch('astropy.table.Table.read')
    h.return_value = hh
    h.data_size = data_size
    return h


def test_sdss_score():
    pass


def test_window_read_no_photo_resolve(monkeypatch):
    monkeypatch.delenv('PHOTO_RESOLVE', raising=False)
    with pytest.raises(PhotoopException) as e:
        window_read()
    assert (str(e.value) ==
            'You have not set the environment variable PHOTO_RESOLVE!')


def test_window_read_all(monkeypatch, mocker, table_read):
    fake_dir = os.path.join('another', 'fake', 'directory')
    monkeypatch.setenv('PHOTO_RESOLVE', fake_dir)
    r = window_read(flist=True, rescore=False, blist=True,
                    bcaps=True, balkans=True, findx=True, bindx=True)
    assert table_read.call_count == 5
    table_read.assert_any_call(os.path.join(fake_dir, 'window_flist.fits'), hdu=1)
    table_read.assert_any_call(os.path.join(fake_dir, 'window_blist.fits'), hdu=1)
    table_read.assert_any_call(os.path.join(fake_dir, 'window_bcaps.fits'), hdu=1)
    table_read.assert_any_call(os.path.join(fake_dir, 'window_findx.fits'), hdu=1)
    table_read.assert_any_call(os.path.join(fake_dir, 'window_bindx.fits'), hdu=1)
    for k in ('flist', 'blist', 'bcaps', 'findx', 'bindx', 'balkans'):
        assert k in r
    for k in ('x', 'cm', 'use_caps'):
        assert hasattr(r['balkans'], k)


def test_window_read_balkans_only(monkeypatch, mocker, table_read):
    fake_dir = os.path.join('another', 'fake', 'directory')
    monkeypatch.setenv('PHOTO_RESOLVE', fake_dir)
    r = window_read(flist=True, rescore=False, balkans=True)
    assert table_read.call_count == 3
    table_read.assert_any_call(os.path.join(fake_dir, 'window_flist.fits'), hdu=1)
    table_read.assert_any_call(os.path.join(fake_dir, 'window_blist.fits'), hdu=1)
    table_read.assert_any_call(os.path.join(fake_dir, 'window_bcaps.fits'), hdu=1)
    for k in ('flist', 'balkans'):
        assert k in r
    assert isinstance(r['balkans'], FITS_polygon)


def test_window_read_rescore(monkeypatch, mocker, table_read):
    fake_dir = os.path.join('another', 'fake', 'directory')
    monkeypatch.setenv('PHOTO_RESOLVE', fake_dir)
    e = mocker.patch('os.path.exists')
    e.return_value = False
    w = mocker.patch('pydl.photoop.window.window_score')
    r = window_read(flist=True, rescore=True)
    assert table_read.call_count == 1
    table_read.assert_any_call(os.path.join(fake_dir, 'window_flist_rescore.fits'), hdu=1)
    # e.assert_called_once_with(os.path.join(fake_dir, 'window_flist_rescore.fits'))
    w.assert_called_once_with(rescore=True)
    for k in ('flist',):
        assert k in r


def test_window_score_no_photo_calib(monkeypatch):
    monkeypatch.delenv('PHOTO_CALIB', raising=False)
    with pytest.raises(PhotoopException) as e:
        window_score()
    assert (str(e.value) ==
            'You have not set the environment variable PHOTO_CALIB!')


def test_window_score_no_photo_resolve(monkeypatch):
    monkeypatch.setenv('PHOTO_CALIB', '/fake/directory')
    monkeypatch.delenv('PHOTO_RESOLVE', raising=False)
    with pytest.raises(PhotoopException) as e:
        window_score()
    assert (str(e.value) ==
            'You have not set the environment variable PHOTO_RESOLVE!')


def test_window_score_no_window_flist(monkeypatch):
    monkeypatch.setenv('PHOTO_CALIB', '/fake/directory')
    monkeypatch.setenv('PHOTO_RESOLVE', '/another/fake/directory')
    with pytest.raises(PhotoopException) as e:
        window_score()
    assert str(e.value) == 'Unable to read FLIST file.'


def test_window_score_fits_open(monkeypatch, mocker, fits_open):
    fake_dir = os.path.join('another', 'fake', 'directory')
    monkeypatch.setenv('PHOTO_CALIB', '/fake/directory')
    monkeypatch.setenv('PHOTO_RESOLVE', fake_dir)
    s = mocker.patch('pydl.photoop.window.sdss_score')
    s.return_value = np.zeros((fits_open.data_size,), dtype=np.int16)
    window_score()
    fits_open.assert_called_once_with(os.path.join(fake_dir, 'window_flist.fits'),
                                      mode='update')
    s.assert_called_once_with(fits_open.return_value)


def test_window_score_fits_open_rescore(monkeypatch, mocker, fits_open):
    fake_dir = os.path.join('another', 'fake', 'directory')
    monkeypatch.setenv('PHOTO_CALIB', '/fake/directory')
    monkeypatch.setenv('PHOTO_RESOLVE', fake_dir)
    s = mocker.patch('pydl.photoop.window.sdss_score')
    s.return_value = np.zeros((fits_open.data_size,), dtype=np.int16)
    window_score(rescore=True)
    assert s.call_count == 1
    fits_open.assert_called_once_with(os.path.join(fake_dir, 'window_flist.fits'),
                                      mode='readonly')
    fits_open.return_value.writeto.assert_called_once_with(os.path.join(fake_dir, 'window_flist_rescore.fits'))
    s.assert_called_once_with(fits_open.return_value)
